import os
import shutil
import click
from .generate import main as generate_main


@click.group()
def cli():  # pragma: no cover
    pass


@click.command()
@click.argument("file")
@click.option("-O", "--outdir", default="./out/", help="The out directory for the generated CSVs")
def generate(file: str, outdir: str):
    if not os.path.isfile(file):
        raise click.UsageError(f"File does not exist: '{file}'")

    # check if the 'out'-directory already exists
    if os.path.isdir(outdir):
        click.confirm(f"The output directory '{outdir}' and ALL ITS CONTENTS will be REMOVED! Continue?",
                      default=False, abort=True, prompt_suffix=": ", show_default=True, err=False)

        shutil.rmtree(outdir)
    # create the directory
    try:
        os.mkdir(outdir)
    except OSError as e:
        raise click.UsageError(e)

    generate_main(file, outdir)


cli.add_command(generate)

if __name__ == '__main__':  # pragma: no cover
    cli()



from typing import Dict

from .schema import SchemaFieldGenerator
from .base import FieldGenerator
from .number import NumberFieldGenerator
from .date import DateFieldGenerator


generators: Dict[str, FieldGenerator] = {
    "number": NumberFieldGenerator,
    "date": DateFieldGenerator
}


def getGeneratorFromFieldType(field_type: str, field_options: Dict[str, str]) -> FieldGenerator:
    Generator = generators.get(field_type, SchemaFieldGenerator)

    return Generator(field_options, field_type)



from schema import Schema

from .base import FieldGenerator


class SchemaFieldGenerator(FieldGenerator):
    option_schema = Schema({})

    def validate(self):
        self.option_schema.validate(self.options)

    def generate(self, res):
        schema = res.get(self.type)

        if schema is None:
            raise Exception(f"Schema '{self.type}' is not generated yet")

        # return id of latest schema
        return schema[1][-1][0][0]



from random import random
from types import FunctionType
from typing import Any, Dict, Tuple, List
import os
import csv

from .field_generators.base import FieldGenerator
from .config import loadConfig


def main(file: str, outdir: str):
    """The main entry point for the generate command

        Params:
            file (str): the validated file path
            outdir (str): the path to the 'out'-directory, where data should be stored
    """
    config = loadConfig(file)
    data = generateByConfig(config)

    # write data to disk
    for key in data.keys():
        path = os.path.join(outdir, f"{key}.csv")

        with open(path, "w", newline='') as file:
            writer = csv.writer(file)
            schema = data.get(key)

            # write header
            writer.writerow(schema[0])
            # write content
            writer.writerows(schema[1])


def generateByConfig(config: Dict) -> Dict[str, Tuple[Tuple[str], List[List]]]:
    """Generates the data based on the configuration file

        Params:
            config (Dict): the validated configuration file

        Returns:
            data (Dict): the generated data by name and rows
    """

    res: Dict[str, Tuple[Tuple, List[List]]] = {}

    iter_amount: int = config.get("amount", 1)
    steps: Dict = config.get("steps")
    initial_step: str = config.get("initial")
    schemas: Dict[str, Dict[str, FieldGenerator]] = config.get("schemas")

    for _ in range(iter_amount):

        # get initial step
        step: Dict = steps.get(initial_step)
        iter_res: Dict[str, Tuple[Tuple,
                                  List[List[Tuple[Any, FunctionType]]]]] = {}

        while step is not None:
            # get schema data
            generate: str | Dict = step.get("generate")

            if type(generate) is str:
                schema_name = generate
            else:
                rand = random()
                total = 0
                schema_name = None

                for (key, value) in generate.items():
                    if (value + total) >= rand:
                        schema_name = key
                        break
                    total += value

            if schema_name is None:
                step = steps.get(step.get("next"))
                continue

            schema: Dict[str, FieldGenerator] = schemas.get(schema_name)
            header = ("id", *schema.keys())

            # initialize schema entry
            if iter_res.get(schema_name) is None:
                iter_res[schema_name] = (header, [])
            # initialize schema values with id
            id = 1 if res.get(schema_name) is None else len(
                res[schema_name][1]) + 1
            schema_values: List[Tuple[Any, FunctionType]] = [(id, str)]

            # generate values
            for field in schema.values():
                if isinstance(field, FieldGenerator):
                    value = field.generate(iter_res)
                    schema_values.append((value, field.val_to_str))
                else:
                    raise Exception(f"Unknown field:\n{field}")

            iter_res[schema_name][1].append(schema_values)

            # advance to next step
            step = steps.get(step.get("next"))

        # add iteration data to complete data
        for key, value in iter_res.items():
            if res.get(key) is None:
                res[key] = (value[0], [list(
                    map(lambda gen: gen[1](gen[0]), *value[1]))])
            else:
                res[key][1].append(map(lambda gen: gen[1](gen[0]), *value[1]))

    return res



from datetime import date, timedelta
from schema import Or, Schema, Use
from random import randint

from .reference import to_ref
from .base import FieldGenerator


class DateFieldGenerator(FieldGenerator):
    option_schema = Schema({
        "min": Or(Use(to_ref), Use(date.fromisoformat)),
        "max": Or(Use(to_ref), Use(date.fromisoformat))
    })

    def validate(self):
        self.options = self.option_schema.validate(self.options)

        # if self.options["max"] < self.options["min"]:
        #     raise SchemaError(
        #         f"'max' ({self.options['max']}) must be equal or larger than 'min' ({self.options['min']})")

    def generate(self, res={}):
        options = self.prepare_options(res)
        min = options["min"]
        max = options["max"]

        day_difference = (max - min).days

        return min + timedelta(days=randint(0, day_difference))



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Tuple

from .reference import ReferenceFieldGenerator


class FieldGenerator(object, metaclass=ABCMeta):
    def __init__(self, options: Dict[str, str] = {}, type: str = None):
        super(FieldGenerator, self).__init__()
        self.options = options
        self.type = type
        self.validate()

    @abstractmethod
    def validate(self):
        raise NotImplementedError()

    @abstractmethod
    def generate(self, res: Dict[str, Tuple[Tuple, List[List]]] = {}):
        raise NotImplementedError()

    def val_to_str(self, value: Any) -> str:
        return str(value)

    def prepare_options(self, res: Dict[str, Tuple[Tuple, List[List]]]):
        options = self.options.copy()
        for (key, value) in self.options.items():
            if not isinstance(value, ReferenceFieldGenerator):
                continue

            schema = res[value.type]
            index = schema[0].index(value.options["field_attr"])

            options[key] = schema[1][-1][index][0]

        return options



"""
This is the "load_config" module.

The load config helps to load and validate the sledo configuration files.
"""

from schema import And, Or, Schema, Optional, SchemaError, Use
from typing import Dict
import yaml

from .field_generators.base import FieldGenerator
from .field_generators.reference import ReferenceFieldGenerator
from .field_generators.schema import SchemaFieldGenerator
from .field_generators import getGeneratorFromFieldType


class ConfigurationSchema(Schema):
    def validate(self, data, _is_config_schema=True):
        data = super(ConfigurationSchema, self).validate(
            data, _is_config_schema=False)

        if not _is_config_schema:
            return data

        steps: Dict[str, Dict[str, str]] = data["steps"]
        schemas: Dict[str, Dict[str, FieldGenerator]] = data["schemas"]

        # validate initial step
        if steps.get(data["initial"]) is None:
            raise SchemaError(f"Missing step: '{data['initial']}'")

        # validate next steps
        for (key, step) in steps.items():
            next = step.get("next")
            if next is not None and steps.get(next) is None:
                raise SchemaError(f"Missing step: '{next}'")

            generate: str | Dict = step.get("generate")

            if type(generate) is str:
                if schemas.get(generate) is None:
                    raise SchemaError(f"Missing schema: '{generate}'")
            else:
                total_prob = 0
                for (schema, prob) in generate.items():
                    if schemas.get(schema) is None:
                        raise SchemaError(f"Missing schema: '{schema}'")

                    total_prob += prob

                if total_prob > 1:
                    raise SchemaError(
                        f"The total probability must not be more than 1 at step: '{key}'")

        # validate properties
        for schema in schemas.values():
            for value in schema.values():
                if isinstance(value, SchemaFieldGenerator):
                    # check if Schema exists
                    ref = schemas.get(value.type)
                    if ref is None:
                        raise SchemaError(f"Missing schema: '{value.type}'")
                elif isinstance(value, FieldGenerator):
                    for value in value.options.values():
                        if isinstance(value, ReferenceFieldGenerator):
                            ref_field = schema.get(value.options["field"])
                            if ref_field is None:
                                raise SchemaError(
                                    f"Missing attribute: '{value.options['field']}'")

                            if not isinstance(ref_field, SchemaFieldGenerator):
                                raise SchemaError(
                                    f"Attribute '{value.options['field']}' must be a schema reference!")

                            value.type = ref_field.type
                            ref_schema = schemas.get(ref_field.type)
                            ref_schema_field = ref_schema.get(
                                value.options["field_attr"])

                            if ref_schema_field is None:
                                raise SchemaError(
                                    f"Attribute '{value.options['field_attr']}' does not exist in schema {ref_field.type}")
        return data


def to_generator(raw: str | dict):
    is_str = type(raw) is str
    field_type = raw if is_str else raw["type"]
    field_options = {} if is_str else {
        k: raw[k] for k in raw if k != "type"
    }

    generator = getGeneratorFromFieldType(field_type, field_options)

    return generator


# The schema of a valid configuration file
schema = ConfigurationSchema({
    'initial': str,
    'amount': Use(int),
    'steps': {
        str: {
            "generate": Or(str, {
                str: And(Use(float), lambda x: 0 <= x <= 1)
            }),
            Optional("next"): str
        }
    },
    'schemas': {
        str: {
            str: And(Or(str, {
                "type": str,
                str: str
            }), Use(to_generator))
        }
    }
})


def validateConfig(config: Dict):
    return schema.validate(config)


def loadConfig(file: str) -> Dict:
    """
    Loads the configuration file and validates its contents

        Params:
            file (str): the file path of the configuration file

        Returns:
            config (Dict): the loaded configuration file
    """

    # load and parse the configuration file
    with open(file) as f:
        config: Dict = yaml.load(f, Loader=yaml.BaseLoader)

        # validate and transform the loaded configuration file
        return validateConfig(config)
